// sertransplg.cpp : Defines the entry point for the DLL application.
//

#include "windows.h"
#include "fsplugin.h"
#include "utils.h"
#include "resource.h"
#include "webdav.h"
#include "davfunc.h"
#include "multiserver.h"

HINSTANCE hinst;
HWND hWndMain=NULL;

#define defininame "tcwebdav.ini"
char inifilename[MAX_PATH]=defininame;
char pluginname[]="WebDAV";
char defrootname[]="WebDAV";

char s_f7newconnection[32];

BOOL disablereading=false;   // disable reading of subdirs to delete whole drives
BOOL freportconnect=true;    // report connect to caller only on first connect

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
                )
{
	if (ul_reason_for_call==DLL_PROCESS_ATTACH) {
		hinst=(HINSTANCE)hModule;
		LoadString(hinst, IDS_F7NEW, s_f7newconnection, sizeof(s_f7newconnection));
	} else if (ul_reason_for_call==DLL_PROCESS_DETACH) {
	}
	return TRUE;
}

int PluginNumber;
tProgressProc ProgressProc;
tLogProc LogProc;
tRequestProc RequestProc;

DWORD lastpercenttime=0;
int lastpercent;

BOOL MessageLoop(void)
{
	BOOL aborted=false;
    if (ProgressProc && abs(GetCurrentTime()-lastpercenttime)>250) {
        // important: also call AFTER soft_aborted is true!!!
        aborted=0!=ProgressProc(PluginNumber,NULL,NULL,lastpercent);
        // allow abort with Escape when there is no progress dialog!
        lastpercenttime=GetCurrentTime();
    }
    return aborted;
}

void ShowStatus(char* status)
{
    if (LogProc)
        LogProc(PluginNumber,MSGTYPE_DETAILS,status);
}

BOOL UpdatePercentBar(int percent)
{
    lastpercent=percent;  // used for MessageLoop below

    return MessageLoop();  // This actually sets the percent bar!
}

SERVERID GetServerIdAndRelativePathFromPath(char* Path,char* RelativePath,int maxlen)
{
	char DisplayName[3*MAX_PATH];
	GetDisplayNameFromPath(Path,DisplayName,sizeof(DisplayName)-1);
	SERVERID serverid=GetServerIdFromName(DisplayName);
	if (serverid) {
		WebDavGetRelativePath(serverid,RelativePath,maxlen);
		char* p=Path;
		while (p[0]=='\\' || p[0]=='/')  // skip initial slash
			p++;
		while (p[0]!=0 && p[0]!='\\' && p[0]!='/') // skip path
			p++;
		strlcat(RelativePath,p,maxlen);
		if (RelativePath[0]==0)
			strcpy(RelativePath,"\\");
	} else if (maxlen)
		strcpy(RelativePath,"\\");
	return serverid;
}

int __stdcall FsInit(int PluginNr,tProgressProc pProgressProc,tLogProc pLogProc,tRequestProc pRequestProc)
{
	ProgressProc=pProgressProc;
	LogProc=pLogProc;
	RequestProc=pRequestProc;
	PluginNumber=PluginNr;
	return 0;
}

typedef struct {
   void* davdataptr;
   SERVERID serverid;
   SERVERHANDLE rootfindhandle;
   BOOL rootfindfirst;
} tLastFindStuct,*pLastFindStuct;

BOOL __stdcall FsDisconnect(char* DisconnectRoot)
{
	char DisplayName[3*MAX_PATH];
	GetDisplayNameFromPath(DisconnectRoot,DisplayName,sizeof(DisplayName)-1);
	SERVERID serverid=GetServerIdFromName(DisplayName);
	if (serverid) {
		char connbuf[MAX_PATH];
		strcpy(connbuf,"DISCONNECT \\");
		strlcat(connbuf,DisplayName,sizeof(connbuf)-1);
		LogProc(PluginNumber,MSGTYPE_DISCONNECT,connbuf);
		WebDavCloseConnection(serverid);
		SetServerIdForName(DisplayName,NULL); // this frees it too!
	}
	return TRUE;
}

HANDLE __stdcall FsFindFirst(char* Path,WIN32_FIND_DATA *FindData)
{
	char remotedir[3*MAX_PATH],DisplayName[MAX_PATH];
	pLastFindStuct lf;

	void* davdataptr=NULL;
	BOOL wasconnected=true;

	if (strcmp(Path,"\\")==0) {  // in the root!
		char s_helptext[256];
		LoadString(hinst, IDS_HELPTEXT, s_helptext, sizeof(s_helptext));
		LoadServersFromIni(inifilename,s_quickconnect);
		memset(FindData,0,sizeof(WIN32_FIND_DATA));

		strcpy(FindData->cFileName,s_f7newconnection);
		FindData->dwFileAttributes=0;
		FindData->ftLastWriteTime.dwHighDateTime=0xFFFFFFFF;
		FindData->ftLastWriteTime.dwLowDateTime=0xFFFFFFFE;
		FindData->nFileSizeLow=strlen(s_helptext);
		lf=(pLastFindStuct)malloc(sizeof(tLastFindStuct));
		memset(lf,0,sizeof(tLastFindStuct));
		lf->rootfindfirst=true;
		return lf;
	} else {
		// load server list if user connects directly via URL
		LoadServersFromIni(inifilename,s_quickconnect);
		// only disable the reading within a server!
		if (disablereading) {
			SetLastError(ERROR_NO_MORE_FILES);
			return INVALID_HANDLE_VALUE;
		}

		GetDisplayNameFromPath(Path,DisplayName,sizeof(DisplayName)-1);
		SERVERID serverid=GetServerIdFromName(DisplayName);
		if (serverid==NULL) {
			wasconnected=false;
			serverid=WebDavConnectToServer(DisplayName,inifilename);
			if (serverid)
				SetServerIdForName(DisplayName,serverid);
			else {
				SetLastError(ERROR_PATH_NOT_FOUND);
				return INVALID_HANDLE_VALUE;
			}
		}
		// we are connected to server DisplayName now!

		if (ProgressProc(PluginNumber,Path,"temp",0)) {
			SetLastError(ERROR_PATH_NOT_FOUND);
			return INVALID_HANDLE_VALUE;
		}

		memset(FindData,0,sizeof(WIN32_FIND_DATA));

		GetServerIdAndRelativePathFromPath(Path,remotedir,sizeof(remotedir)-1);

		// Retrieve the directory
		if (DAV_OK!=WebDavFindFirstFile(serverid,remotedir,&davdataptr)) {
			if (!wasconnected) {  // initial connect failed
				WebDavCloseConnection(serverid);
				SetServerIdForName(DisplayName,NULL); // this frees it too!
				freportconnect=false;
			}
			SetLastError(ERROR_PATH_NOT_FOUND);
			return INVALID_HANDLE_VALUE;
		}

		if (DAV_OK==WebDavFindNextFile(serverid,davdataptr,FindData)) {
			lf=(pLastFindStuct)malloc(sizeof(tLastFindStuct));
			memset(lf,0,sizeof(tLastFindStuct));
			lf->davdataptr=davdataptr;
			lf->serverid=serverid;
			return (HANDLE)lf;
		} else {
			WebDavFindClose(serverid,davdataptr);
			SetLastError(ERROR_NO_MORE_FILES);
			return INVALID_HANDLE_VALUE;
		}
	}
	return INVALID_HANDLE_VALUE;
}

BOOL __stdcall FsFindNext(HANDLE Hdl,WIN32_FIND_DATA *FindData)
{
	pLastFindStuct lf;

	if ((int)Hdl==1)
		return false;

	lf=(pLastFindStuct)Hdl;
	if (lf!=INVALID_HANDLE_VALUE) {
		if (lf->rootfindfirst) {
			SERVERHANDLE hdl=FindFirstServer(FindData->cFileName,sizeof(FindData->cFileName)-1);
			if (hdl) {
				lf->rootfindhandle=hdl;
				lf->rootfindfirst=false;
				FindData->dwFileAttributes=FILE_ATTRIBUTE_DIRECTORY;
				FindData->ftLastWriteTime.dwHighDateTime=0xFFFFFFFF;
				FindData->ftLastWriteTime.dwLowDateTime=0xFFFFFFFE;
				FindData->nFileSizeLow=0;
				return true;
			} else
				return false;
		} else if (lf->rootfindhandle) {
			lf->rootfindhandle=FindNextServer(lf->rootfindhandle,FindData->cFileName,sizeof(FindData->cFileName)-1);
			return lf->rootfindhandle!=NULL;
		} else
			return WebDavFindNextFile(lf->serverid,lf->davdataptr,FindData)==DAV_OK;
	}
	return false;
}

int __stdcall FsFindClose(HANDLE Hdl)
{
   if (Hdl==INVALID_HANDLE_VALUE)
      return 0;
   pLastFindStuct lf;
   lf=(pLastFindStuct)Hdl;
   if (lf->davdataptr) {
	   WebDavFindClose(lf->serverid,lf->davdataptr);
	   lf->davdataptr=NULL;
   }
   free(lf);
   return 0;
}

BOOL __stdcall FsMkDir(char* Path)
{
	char* p=strchr(Path+1,'\\');
	if (p) {
		char remotedir[3*MAX_PATH];
		SERVERID serverid=GetServerIdAndRelativePathFromPath(Path,remotedir,sizeof(remotedir)-1);
		if (serverid)
			return WebDavCreateDirectory(serverid,remotedir)==DAV_OK;
		else
			return false;
	} else {  // new connection
		if (strcmp(Path+1,s_quickconnect)!=0) {
			LoadServersFromIni(inifilename,s_quickconnect);
			if (WebDavConfigureServer(Path+1,inifilename)) {
				LoadServersFromIni(inifilename,s_quickconnect);
				return true;
			} else
				return false;
		} else
			return false;
	}
}

int __stdcall FsExecuteFile(HWND MainWin,char* RemoteName,char* Verb)
{
	if (stricmp(Verb,"open")==0) {
		return FS_EXEC_YOURSELF;
	} else if (stricmp(Verb,"properties")==0) {
		if (RemoteName[1] && strchr(RemoteName+1,'\\')==0 && strcmp(RemoteName+1,s_f7newconnection)!=0) {
			if (WebDavConfigureServer(RemoteName+1,inifilename)) {
				LoadServersFromIni(inifilename,s_quickconnect);
			}
			return FS_EXEC_OK;
		} else
			return FS_EXEC_ERROR;
	} else
		return FS_EXEC_ERROR;
}

int __stdcall FsRenMovFile(char* OldName,char* NewName,BOOL Move,BOOL OverWrite,RemoteInfoStruct* ri)
{
	lastpercent=0;
	char olddir[3*MAX_PATH],newdir[3*MAX_PATH];

	// Rename or copy a server?
	char* p1=strchr(OldName+1,'\\');
	char* p2=strchr(NewName+1,'\\');
	if (p1==NULL && p2==NULL) {
		switch (CopyMoveServerInIni(OldName+1,NewName+1,Move,OverWrite,inifilename)) {
			case 0:return FS_FILE_OK; break;
			case 1:return FS_FILE_EXISTS; break;
		}
		return FS_FILE_NOTFOUND;

	}

	SERVERID serverid1=GetServerIdAndRelativePathFromPath(OldName,olddir,sizeof(olddir)-1);
	SERVERID serverid2=GetServerIdAndRelativePathFromPath(NewName,newdir,sizeof(newdir)-1);

	// must be on same server!
	if (serverid1!=serverid2 || serverid1==NULL)
		return FS_FILE_NOTFOUND;
	BOOL isdir=false;
	if (ri)
		isdir=(ri->Attr & FILE_ATTRIBUTE_DIRECTORY)!=0;

	switch (WebDavRenameMoveFile(serverid1,olddir,newdir,Move,OverWrite,isdir)) {
	case DAV_OK:
		return FS_FILE_OK;
	case DAV_EXISTS:
		return FS_FILE_EXISTS;
	default:
		return FS_FILE_NOTFOUND;
	}
}

BOOL FileExists(char* LocalName)
{
	WIN32_FIND_DATA s;
	HANDLE findhandle;
	findhandle=FindFirstFile(LocalName,&s);
	if (findhandle==INVALID_HANDLE_VALUE)
		return false;
	else {
		FindClose(findhandle);
		return true;
	}
}

void RemoveInalidChars(char* p)
{
   while (p[0]) {
      if ((unsigned char)(p[0])<32)
         p[0]=' ';
      else if (p[0]==':' || p[0]=='|' || p[0]=='*' || p[0]=='?' || p[0]=='\\' || p[0]=='/' || p[0]=='"')
         p[0]='_';
      p++;
   }
}

int __stdcall FsGetFile(char* RemoteName,char* LocalName,int CopyFlags,RemoteInfoStruct* ri)
{
	int err;
	BOOL OverWrite,Resume,Move;

	lastpercent=0;
	OverWrite=CopyFlags & FS_COPYFLAGS_OVERWRITE;
	Resume=CopyFlags & FS_COPYFLAGS_RESUME;
	Move=CopyFlags & FS_COPYFLAGS_MOVE;

	if (strlen(RemoteName)<3)
		return FS_FILE_NOTFOUND;

	if (strcmp(RemoteName+1,s_f7newconnection)==0) {
		HANDLE houtfile=CreateFile(LocalName,GENERIC_WRITE,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,
			OverWrite ? CREATE_ALWAYS : CREATE_NEW,
			FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,NULL);
		if (houtfile!=INVALID_HANDLE_VALUE) {
			DWORD written;
			char s_helptext[256];
			LoadString(hinst, IDS_HELPTEXT, s_helptext, sizeof(s_helptext));
			BOOL ret=WriteFile(houtfile,s_helptext,strlen(s_helptext),&written,NULL);
			CloseHandle(houtfile);
			if (ret)
				return FS_FILE_OK;
		} else
			if (OverWrite)
				return FS_FILE_EXISTS;
		return FS_FILE_WRITEERROR;
	}

	char* p=strrchr(LocalName,'\\');
	if (p)
		RemoveInalidChars(p+1);  // Changes the name passed in!

	err=ProgressProc(PluginNumber,RemoteName,LocalName,0);
	if (err)
		return FS_FILE_USERABORT;
	if (OverWrite)
		DeleteFile(LocalName);
	else {
		if (!Resume && FileExists(LocalName))
			return FS_FILE_EXISTSRESUMEALLOWED;
	}

	_int64 filesize=(((_int64)ri->SizeHigh)<<32)+ri->SizeLow;

	char remotedir[3*MAX_PATH];
	
	SERVERID serverid=GetServerIdAndRelativePathFromPath(RemoteName,remotedir,sizeof(remotedir)-1);
	if (serverid==NULL)
		return FS_FILE_READERROR;

	switch (WebDavDownloadFile(serverid,remotedir,LocalName,true,filesize,&ri->LastWriteTime,Resume)) {
		case DAV_OK:return FS_FILE_OK;
		case DAV_EXISTS:return FS_FILE_EXISTS;
		case DAV_READFAILED:return FS_FILE_READERROR;
		case DAV_WRITEFAILED:return FS_FILE_WRITEERROR;
		case DAV_ABORT:return FS_FILE_USERABORT;
	}
	return FS_FILE_OK;
}


int __stdcall FsPutFile(char* LocalName,char* RemoteName,int CopyFlags)
{
	int err;
	DWORD resumeat;
	BOOL OverWrite,Resume,Move;

	lastpercent=0;
	OverWrite=CopyFlags & FS_COPYFLAGS_OVERWRITE;
	Resume=CopyFlags & FS_COPYFLAGS_RESUME;
	Move=CopyFlags & FS_COPYFLAGS_MOVE;

	// WebDAV auto-overwrites files -> return error if file exists
	if ((CopyFlags & (FS_COPYFLAGS_EXISTS_SAMECASE | FS_COPYFLAGS_EXISTS_DIFFERENTCASE)) &&
		!(OverWrite | Resume))
		return FS_FILE_EXISTS;   // resume not supported by HTTP PUT!

	if (strlen(RemoteName)<3)
		return FS_FILE_WRITEERROR;

	err=ProgressProc(PluginNumber,LocalName,RemoteName,0);
	if (err)
		return FS_FILE_USERABORT;

	char remotedir[3*MAX_PATH];
	
	SERVERID serverid=GetServerIdAndRelativePathFromPath(RemoteName,remotedir,sizeof(remotedir)-1);
	if (serverid==NULL)
		return FS_FILE_READERROR;

	if (OverWrite && !Resume && WebDavDeleteBeforeUpload(serverid))  // needed for 1&1 webdav server!
		WebDavDeleteFile(serverid,remotedir,false);

	resumeat=0;
	if (Resume)
		return FS_FILE_NOTSUPPORTED;

	switch (WebDavUploadFile(serverid,LocalName,remotedir,resumeat)) {
		case DAV_OK:return FS_FILE_OK;
		case DAV_EXISTS:return FS_FILE_EXISTSRESUMEALLOWED;
		case DAV_READFAILED:return FS_FILE_READERROR;
		case DAV_WRITEFAILED:return FS_FILE_WRITEERROR;
		case DAV_ABORT:return FS_FILE_USERABORT;
	}
	return FS_FILE_NOTFOUND;
}

BOOL __stdcall FsDeleteFile(char* RemoteName)
{
	if (strlen(RemoteName)<3)
		return false;

	char remotedir[3*MAX_PATH];
	SERVERID serverid=GetServerIdAndRelativePathFromPath(RemoteName,remotedir,sizeof(remotedir)-1);
	if (serverid==NULL)
		return false;
	
	lastpercent=0;
	return WebDavDeleteFile(serverid,remotedir,false)==DAV_OK;
}

BOOL __stdcall FsRemoveDir(char* RemoteName)
{
	if (strlen(RemoteName)<1)
		return false;

	char* p=strchr(RemoteName+1,'\\');
	if (p) {
		char remotedir[3*MAX_PATH];
		SERVERID serverid=GetServerIdAndRelativePathFromPath(RemoteName,remotedir,sizeof(remotedir)-1);
		if (serverid==NULL)
			return FS_FILE_READERROR;
		lastpercent=0;

        return (WebDavDeleteFile(serverid,remotedir,true)==DAV_OK);
	} else {	// delete server
		return DeleteServerFromIni(RemoteName+1,inifilename);
	}
}

BOOL __stdcall FsSetAttr(char* RemoteName,int NewAttr)
{
	char remotedir[3*MAX_PATH];
	SERVERID serverid=GetServerIdAndRelativePathFromPath(RemoteName,remotedir,sizeof(remotedir)-1);
	if (serverid==NULL)
		return FS_FILE_READERROR;
	
	lastpercent=0;
	return WebDavSetAttr(serverid,remotedir,NewAttr)==DAV_OK;
}

BOOL __stdcall FsSetTime(char* RemoteName,FILETIME *CreationTime,
      FILETIME *LastAccessTime,FILETIME *LastWriteTime)
{
	char remotedir[3*MAX_PATH];
	SERVERID serverid=GetServerIdAndRelativePathFromPath(RemoteName,remotedir,sizeof(remotedir)-1);
	if (serverid==NULL)
		return FS_FILE_READERROR;

	lastpercent=0;
	return WebDavSetDateTime(serverid,remotedir,LastWriteTime)==DAV_OK;
}

void __stdcall FsStatusInfo(char* RemoteDir,int InfoStartEnd,int InfoOperation)
{
	if (strlen(RemoteDir)<2)
		if (InfoOperation==FS_STATUS_OP_DELETE || InfoOperation==FS_STATUS_OP_RENMOV_MULTI)
			if (InfoStartEnd==FS_STATUS_START)
				disablereading=true;
			else
				disablereading=false;
}

void __stdcall FsGetDefRootName(char* DefRootName,int maxlen)
{
   strlcpy(DefRootName,defrootname,maxlen);
}

// use default location, but our own ini file name!
void __stdcall FsSetDefaultParams(FsDefaultParamStruct* dps)
{
	strlcpy(inifilename,dps->DefaultIniName,MAX_PATH-1);
	char* p=strrchr(inifilename,'\\');
	if (p)
		p[1]=0;
	else
		inifilename[0]=0;
	strlcat(inifilename,defininame,sizeof(inifilename)-1);
}

